/*
 * This file is part of MPlayer.
 *
 * MPlayer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MPlayer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with MPlayer; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "img_format.h"
#include "mp_image.h"
#include "mp_msg.h"
#include "vf.h"

#include "libvo/fastmemcpy.h"
#include "libavutil/common.h"

struct vf_priv_s {
    int w;  //1=lr,2=rl,3=uo,4=ou,5=auto(todo), 0 - none
	int d;  //picture offsets
	int show_info;
};

static int
config(struct vf_instance* vf,
       int width, int height, int d_width, int d_height,
       unsigned int flags, unsigned int outfmt)
{
  if(vf->priv->w<0 || vf->priv->w>5) 
  {
    mp_msg(MSGT_VFILTER,MSGL_WARN,"3d checkboard:0-4 !\n");
    return 0;
  }
  mp_msg(MSGT_VFILTER,MSGL_WARN,"3d checkboard init OK!\n");
  return vf_next_config(vf, width, height, d_width, d_height, flags, outfmt);
}

static int
control(struct vf_instance* vf, int request, void *data)
{
    const int *const tmp = data;
    switch(request){
    //todo : swap lr<>rl, ou<>uo
    }
    return vf_next_control(vf, request, data);
    return 0;
}
static int
put_image(struct vf_instance* vf, mp_image_t* mpi, double pts){
    mp_image_t* dmpi;
	unsigned int px,py,pb;
    unsigned int bpp = mpi->bpp / 8;
    unsigned char bx[8000];	//buffer for one line fullhd, up to 2000 pixels (8000/4 for rgb32)
	dmpi = vf_get_image(vf->next, mpi->imgfmt, MP_IMGTYPE_TEMP,
			MP_IMGFLAG_ACCEPT_STRIDE | MP_IMGFLAG_PREFER_ALIGNED_STRIDE,
			mpi->w, mpi->h);

    memcpy_pic(dmpi->planes[0],mpi->planes[0],mpi->w*bpp, mpi->h,
	       dmpi->stride[0],mpi->stride[0]);

    //mp_msg(MSGT_VFILTER,MSGL_INFO, "3dcb: -vf 3dcb=%d \n", vf->priv->w);

    /* Change pixel sort, for example using 3dcb=1 (lr):
    AAAABBBB     ABABABAB
    AAAABBBB  => BABABABA
    AAAABBBB     ABABABAB    
    */

	//lr/rl

	if(vf->priv->w==1 || vf->priv->w==2)
	{
		for( py=0; py<dmpi->h; py++ )
		{
			unsigned char *p = dmpi->planes[0] + py * dmpi->stride[0] ;	//row start
			unsigned char *q = dmpi->planes[0] + py * dmpi->stride[0] + dmpi->w/2*bpp;	//row half
			if((py%2==0 && vf->priv->w==1) || (py%2==1 && vf->priv->w==2))
			{	//swap p q
				p = q;
				q = dmpi->planes[0] + py * dmpi->stride[0];
			}
			q=q+vf->priv->d;
			for( px=0; px<mpi->w*bpp; px+=bpp*2 )	//read one row and pixel mixing
			{
				for( pb=0; pb<bpp; pb++ )	//inside one pixel loop
				{	
					bx[px+pb] = *p++;
					bx[px+bpp+pb] = *q++;
				}
			}

			//write mixed row
			p = dmpi->planes[0] + py * dmpi->stride[0] ;	//row start
			for( px=0; px<mpi->w*bpp; px++ )
			{
				*p++=bx[px];
			}
		}
	}

	//ud/du
	if(vf->priv->w==3 || vf->priv->w==4)
	{
		for( px=0; px<dmpi->w; px++ )
		{
			unsigned char *p = dmpi->planes[0] + px*bpp ;	//row start 1
			unsigned char *q = dmpi->planes[0] + (dmpi->h/2)*bpp*dmpi->w +px*bpp;	//row start 2
			if((px%2==0 && vf->priv->w==3) || (px%2==1 && vf->priv->w==4))
			{
				p=q;
				q=dmpi->planes[0] + px*bpp;
			}
			for( py=0; py<mpi->h/2; py+=1 )	//read one column and pixel mixing
			{
				for( pb=0; pb<bpp; pb++ )	//inside one pixel loop
				{	
					bx[2*py*bpp+pb] = *p++;
					bx[2*py*bpp+bpp+pb] = *q++;
				}
				p=p+bpp*dmpi->w-bpp;
				q=q+bpp*dmpi->w-bpp;
			}

			//write mixed row
			p = dmpi->planes[0] + px*bpp ;	//row start
			for( py=0; py<mpi->h; py++ )
			{
				for( pb=0; pb<bpp; pb++ )	//inside one pixel loop
				{	
					*p++=bx[py*bpp+pb];
				}
				p=p+bpp*dmpi->w-bpp;
			}
		}
	}

    return vf_next_put_image(vf, dmpi, pts);
}

static int query_format(struct vf_instance *vf, unsigned int fmt){
	//only those two tested, becouse it needs full color info for every pixel
    switch(fmt){
    case IMGFMT_BGR24:
    case IMGFMT_BGR32:
	return vf_next_query_format(vf,fmt);
    }
	if(vf->priv->show_info)
	{
		mp_msg(MSGT_VFILTER,MSGL_INFO, "**************************\n3dcb: use rgb24 or rgb32 output, try '-vo x11'\n**************************\n");
		vf->priv->show_info=0;
	}
    return 0;
}

static int
vf_open(vf_instance_t *vf, char *args) {
    vf->config = config;
	vf->query_format=query_format;
    vf->control = control;
    vf->put_image = put_image;
    vf->priv = malloc(sizeof(struct vf_priv_s));
    vf->priv->w = 0;
    vf->priv->d = 0;
	vf->priv->show_info = 1;
    if (args)
  		sscanf(args, "%d:%d", &vf->priv->w,&vf->priv->d);
    return 1;
}

const vf_info_t vf_info_3dcb = {
    "transform RGB/BGR image to 3d checkboard for PDP viewing",
    "3dcb",
    "Krzysiek Sporysz",
    "",
    vf_open,
    NULL
};
